from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram

from biobalm.space_utils import is_subspace
from biobalm.trappist_core import trappist


def expand_minimal_spaces(sd: SuccessionDiagram, size_limit: int | None = None) -> bool:
    """
    See `SuccessionDiagram.expand_minimal_spaces` for documentation.
    """

    minimal_traps = trappist(sd.petri_net, problem="min")

    root = sd.root()

    seen = set([root])

    stack: list[tuple[int, list[int] | None]] = [(root, None)]

    while len(stack) > 0:
        (node, successors) = stack.pop()
        if successors is None:
            # Only allow successor computation if size limit hasn't been exceeded.
            if (size_limit is not None) and (len(sd) >= size_limit):
                # Size limit reached.
                return False

            successors = sd.node_successors(node, compute=True)
            successors = sorted(successors, reverse=True)  # For determinism!
            # (reversed because we explore the list from the back)

        node_space = sd.node_data(node)["space"]

        # Remove all immediate successors that are already visited or those who
        # do not cover any new minimal trap space.
        while len(successors) > 0:
            if successors[-1] in seen:
                successors.pop()
                continue
            if len([s for s in minimal_traps if is_subspace(s, node_space)]) == 0:
                successors.pop()
                continue
            break

        # This node is done and we don't have to push anything onto the stack.
        # In particular, it means that every minimal trap space which is a subset
        # of this node is already in the succession diagram.
        if len(successors) == 0:
            if sd.node_is_minimal(node):
                minimal_traps.remove(sd.node_data(node)["space"])
            continue

        # At this point, we know that `s` is not visited and it contains
        # at least one minimal trap space that does not appear in the
        # succession diagram yet.

        s = successors.pop()
        seen.add(s)
        # Push the node back with the remaining successors.
        stack.append((node, successors))
        # Push the successor onto the stack.
        stack.append((s, None))

    assert len(minimal_traps) == 0
    return True



from __future__ import annotations

import itertools as it
from typing import TYPE_CHECKING, cast

from biobalm.space_utils import percolate_network
from biobalm.types import BooleanSpace
from biobalm.interaction_graph_utils import source_nodes

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram


def expand_source_blocks(
    sd: SuccessionDiagram,
    check_maa: bool = True,
) -> bool:
    """
    Base correctness assumptions:

     - Expanding two minimal blocks is always independent.
     -
    """

    if sd.config["debug"]:
        print(f"Start SD expansion using block decomposition on {sd.network}.")

    root = sd.root()

    current_level: set[int] = set([root])
    next_level: set[int] = set()

    # This already accounts for constant percolation.
    node_space = sd.node_data(root)["space"]

    # find source nodes
    perc_bn = percolate_network(sd.network, node_space)
    sources = source_nodes(perc_bn)

    if sd.config["debug"]:
        print(f" > Computed source/input variable(s): {sources}")

    # get source nodes combinations and expand root node
    if len(sources) != 0:
        # If there are too many source nodes, this can generate an absurdly large SD.
        # This would be a problem even without the SCC expansion, but we can just
        # stop the whole thing faster because we know how many nodes it generates.
        if 2 ** len(sources) > sd.config["max_motifs_per_node"]:
            raise RuntimeError(
                f"Exceeded the maximum amount of stable motifs per node ({sd.config['max_motifs_per_node']}; see `SuccessionDiagramConfiguration.max_motifs_per_node`)."
            )
        else:
            if sd.config["debug"]:
                print(
                    f" > Expanding {len(sources)} source node into {2 ** len(sources)} SD nodes."
                )

        bin_values_iter = it.product(range(2), repeat=len(sources))
        for bin_values in bin_values_iter:
            valuation = cast(BooleanSpace, dict(zip(sources, bin_values)))
            sub_space = node_space | valuation

            next_level.add(sd._ensure_node(root, sub_space))  # type: ignore

        # This makes the root artificially "expanded". Also, there
        # can be no attractors here because we are just fixing the source nodes.
        sd.node_data(root)["expanded"] = True
        sd.node_data(root)["attractor_seeds"] = []
        sd.node_data(root)["attractor_sets"] = []
        current_level = next_level
        next_level = set()

    bfs_depth = 0

    while len(current_level) > 0:
        bfs_depth += 1
        if sd.config["debug"]:
            print(
                f" > Start block expansion of a BFS level {bfs_depth} with {len(current_level)} node(s)."
            )

        for node in sorted(current_level):  # Sorted for determinism
            if sd.node_data(node)["expanded"]:
                # We re-discovered a previously expanded node.
                continue

            # Compute successors as if this was a normal expansion procedure.
            successors = sd.node_successors(node, compute=True)
            # Sort successors to avoid non-determinism.
            successors = sorted(successors)

            if len(successors) == 0:
                # This is a minimal trap space.
                continue

            if len(successors) == 1 and not check_maa:
                # This space is not minimal, but there is no "choice" to make here.
                # We can just continue with that single choice.
                #
                # However, if we are checking for MAAs, we continue as normal, because
                # we need to compute most of the non-trivial results anyway.

                next_level.add(successors[0])
                continue

            node_bn = sd.node_percolated_network(node, compute=True)

            # Split successors into "blocks" based on the regulatory component
            # of the variables fixed by the stable motif.

            # Maps a "block" (bwd-closed set of variables) to a list of node IDs (successor nodes).
            blocks: list[tuple[set[str], list[int]]] = []
            for s in successors:
                motif = sd.edge_stable_motif(node, s, reduced=True)
                motif_block = node_bn.backward_reachable(list(motif.keys()))
                motif_block_names = {node_bn.get_variable_name(v) for v in motif_block}

                found = False
                for block, nodes in blocks:
                    if block == motif_block_names:
                        found = True
                        nodes.append(s)
                        break
                if not found:
                    blocks.append((motif_block_names, [s]))

            if sd.config["debug"]:
                print(
                    f" > [{node}] Computed blocks: {[(len(k), len(v)) for (k, v) in blocks]}"
                )

            # Now remove all non-minimal blocks (minimal in terms of set inclusion).
            # The reason why we are removing those is that they are not independent
            # on the minimal blocks.

            minimal_blocks: list[tuple[set[str], list[int]]] = []
            if len(blocks) > 1:
                for block, nodes in blocks:
                    is_minimal = True
                    for b2, _ in blocks:
                        if b2 < block:
                            is_minimal = False
                            break
                    if is_minimal:
                        minimal_blocks.append((block, nodes))
            else:
                minimal_blocks = blocks

            # Sort minimal blocks by the number of successor nodes (we want to pick
            # the one that leads leads to the minimal SD expansion).
            minimal_blocks = sorted(minimal_blocks, key=lambda x: len(x[1]))

            if sd.config["debug"]:
                print(
                    f" > [{node}] Minimal blocks: {[(len(k), len(v)) for (k, v) in minimal_blocks]}"
                )

            if not check_maa:
                # We will expand all nodes that are in the smallest block.
                to_expand = minimal_blocks[0][1]

                if sd.config["debug"]:
                    print(f" > [{node}] Final block ({len(to_expand)}): {to_expand}")

                next_level = next_level | set(to_expand)
            else:
                # Here, we want to find the smallest block without any MAAs and choose it.
                # If such block does not exist, we expand the whole node, because the MAAs
                # can be either in this node, or in any of the child nodes. If a clean block
                # is found, we know that it is safe to expand it without "missing" any MAAs.
                clean_block_found = False
                for block, block_nodes in minimal_blocks:
                    block_sd = sd.component_subdiagram(list(block), node)

                    # The succession diagram "restricted" to the considered block should have
                    # the same (restricted) successor nodes.
                    assert len(
                        block_sd.node_successors(block_sd.root(), compute=True)
                    ) == len(block_nodes)

                    # We could also consider using `seeds` instead of `candidates` here. Ultimately, this
                    # matters very rarely. The reasoning for why we use `candidates` is that we can (almost)
                    # always guarantee that the expansion is going to finish, even if some nodes do not
                    # have their MAAs eliminated. As such, one can try to use other techniques to disprove
                    # MAAs in the problematic nodes while using the nice properties of the expansion to
                    # still disprove MAAs in the remaining nodes. If we used `seeds`, the expansion could
                    # just get stuck on this node and the "partial" results wouldn't be usable.
                    block_sd_candidates = block_sd.node_attractor_candidates(
                        block_sd.root(), compute=True
                    )
                    if len(block_sd_candidates) == 0:
                        if sd.config["debug"]:
                            print(
                                f" > [{node}] Found clean block with no MAAs ({len(block_nodes)}): {block_nodes}"
                            )
                        clean_block_found = True
                        next_level = next_level | set(block_nodes)
                        sd.node_data(node)["attractor_seeds"] = []
                        sd.node_data(node)["attractor_sets"] = []
                        break
                    else:
                        if sd.config["debug"]:
                            print(
                                f"[{node}] > Found {len(block_sd_candidates)} MAA cnadidates in a block. Delaying expansion."
                            )
                if not clean_block_found:
                    # If all blocks have MAAs, we expand all successors.
                    if sd.config["debug"]:
                        print(
                            f" > [{node}] No clean block found. Expanding all {len(successors)} successors."
                        )
                    next_level = next_level | set(successors)

        current_level = next_level
        next_level = set()

    if sd.config["debug"]:
        print(f" > Block expansion terminated with {len(sd)} node(s).")

    return True



from __future__ import annotations

from typing import TYPE_CHECKING, cast, Literal

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram
    from biodivine_aeon import VariableId

from biodivine_aeon import AsynchronousGraph, ColoredVertexSet, VertexSet
from biobalm.symbolic_utils import state_list_to_bdd
from biobalm.types import BooleanSpace


def compute_attractors_symbolic(
    sd: SuccessionDiagram,
    node_id: int,
    candidate_states: list[BooleanSpace],
    seeds_only: bool = False,
) -> tuple[list[BooleanSpace], list[VertexSet] | None]:
    """
    Uses exhaustive symbolic reachability to eliminate spurious candidate states
    and to compute the exact attractor sets.

    Parameters
    ----------
    sd : SuccessionDiagram
        The succession diagram in question.
    node_id : int
        An ID of an SD node for which we are computing the candidates.
    candidate_states: bool
        The list of candidate states covering the attractors in the
        given node.
    seeds_only: bool
        If `True`, the method can terminate early once it is guaranteed that
        all seeds have been identified. In such case, the list of returned sets
        is `None`.

    Returns
    -------
    tuple[list[BooleanSpace], list[VertexSet] | None]
        The list of attractor seed states and corresponding attractor sets (if computed).
    """

    node_data = sd.node_data(node_id)
    node_space = node_data["space"]
    bn_reduced = sd.node_percolated_network(node_id, compute=True)
    graph_reduced = AsynchronousGraph(bn_reduced)
    symbolic_ctx = graph_reduced.symbolic_context()

    # Ignore variables that are already fixed in the nodes space.
    # These are not used in the reduced network.
    candidate_states_reduced: list[BooleanSpace] = []
    for candidate in candidate_states:
        candidate_reduced: BooleanSpace = {
            k: v for (k, v) in candidate.items() if k not in node_space
        }
        candidate_states_reduced.append(candidate_reduced)

    child_motifs_reduced = []
    if node_data["expanded"]:
        children = sd.node_successors(node_id, compute=False)
        child_motifs_reduced = [
            sd.edge_stable_motif(node_id, s, reduced=True) for s in children
        ]

    child_motifs_bdd = state_list_to_bdd(symbolic_ctx, child_motifs_reduced)
    candidate_bdd = state_list_to_bdd(symbolic_ctx, candidate_states_reduced)

    children_set = ColoredVertexSet(symbolic_ctx, child_motifs_bdd)
    candidate_set = ColoredVertexSet(symbolic_ctx, candidate_bdd)

    avoid = candidate_set.union(children_set)

    if sd.config["debug"]:
        print(
            f"[{node_id}] > Start symbolic seed state identification with {len(candidate_states)} candidates and {avoid} avoid states."
        )

    seeds: list[BooleanSpace] = []
    sets: list[ColoredVertexSet] = []
    for i, candidate in enumerate(candidate_states_reduced):
        is_last = i == len(candidate_states_reduced) - 1
        is_minimal = len(child_motifs_reduced) == 0
        if seeds_only and is_minimal and is_last and len(seeds) == 0:
            # This node is pseudo-minimal, so it must contain at least
            # one attractor seed. If we only care about the attractor seeds,
            # we can thus return the last candidate without checking it.
            if sd.config["debug"]:
                print(
                    f"[{node_id}] > Single seed remaining in a (pseduo) minimal space. Done."
                )
            return ([candidate | node_space], None)
        candidate_singleton = graph_reduced.mk_subspace(candidate)

        avoid = avoid.minus(candidate_singleton)

        closure = symbolic_attractor_test(sd, node_id, graph_reduced, candidate, avoid)

        if closure is None:
            # This candidate can reach someone else in the candidate set,
            # or they can reach one of the child spaces. Hence it is not
            # an attractor.
            continue

        # Otherwise, we have an attractor set and we can save it.
        # It also becomes a part of the avoid set.
        avoid = avoid.union(closure)
        seeds.append(candidate | node_space)
        sets.append(closure)

    if sd.config["debug"]:
        print(f"[{node_id}] > Finished identification with {len(seeds)} seed states.")

    space_symbolic = sd.symbolic.mk_subspace(node_space).vertices()
    sets_converted: list[VertexSet] = []
    for s in sets:
        # Extend the attractor set with fixed nodes from the node space.
        vertices = s.vertices()
        vertices = sd.symbolic.transfer_from(vertices, graph_reduced)
        vertices = vertices.intersect(space_symbolic)
        sets_converted.append(vertices)

    return (seeds, sets_converted)


def symbolic_attractor_test(
    sd: SuccessionDiagram,
    node_id: int,
    graph: AsynchronousGraph,
    pivot: BooleanSpace,
    avoid_set: ColoredVertexSet,
) -> ColoredVertexSet | None:
    """
    Use symbolic reachability with saturation to compute the set of states
    reachable from the given `pivot`, or `None` if the reachability procedure
    intersects with any state in the `avoid` set.

    *The reason why we use `ColoredVertexSet` instead of `VertexSet` is mostly
    a technicality that should be irelevant in biobalm, since we don't allow any
    parameters outside of unknown inputs.*
    """

    if avoid_set.is_empty():
        avoid = None
    else:
        avoid = avoid_set

    reach_set = graph.mk_subspace(pivot)

    # The variables for which we already maintain that `reach_set`
    # contains all reachable states.
    saturated_vars: list[VariableId] = []

    # Variables where `reach_set` differs from the states in `avoid`.
    # We should prioritize updating these variables, because they *need*
    # to be updated if we are ever to reach `avoid`.
    conflict_vars: list[VariableId] = []

    # Populate conflict vars, assuming we have any.
    if avoid is not None:
        avoid_bdd = avoid.to_bdd()
        for name, val in pivot.items():
            var = graph.find_network_variable(name)
            bdd_var = graph.symbolic_context().find_network_bdd_variable(name)
            assert var is not None and bdd_var is not None
            incompatible = avoid_bdd.r_select({bdd_var: cast(Literal[0, 1], 1 - val)})
            if not incompatible.is_false():
                conflict_vars.append(var)
        conflict_vars = sort_variable_list(conflict_vars)

    # Remaining network variables that are still relevant, but may not
    # be necessary to reach `avoid`.
    other_vars: list[VariableId] = [
        x for x in graph.network_variables() if x not in conflict_vars
    ]
    other_vars = sort_variable_list(other_vars)

    if sd.config["debug"]:
        print(
            f"[{node_id}] > Start symbolic reachability with {len(conflict_vars)} conflict variables and {len(other_vars)} other variables."
        )

    all_done = False
    while not all_done:
        all_done = True

        # Saturate reach set with currently selected variables.
        saturation_done = False
        while not saturation_done:
            if avoid is not None and not avoid.intersect(reach_set).is_empty():
                if sd.config["debug"]:
                    print(f"[{node_id}] > Discovered avoid state. Done.")
                return None

            saturation_done = True
            for var in saturated_vars:
                successors = graph.var_post_out(var, reach_set)
                if not successors.is_empty():
                    reach_set = reach_set.union(successors)
                    saturation_done = False
                    if reach_set.symbolic_size() > 100_000 and sd.config["debug"]:
                        print(
                            f"[{node_id}] > Saturation({len(saturated_vars)}) Expanded reach_set: {reach_set}"
                        )
                    break

        # Once saturated, try to expand the saturated
        # collection with either a conflict variable or
        # other variable.

        # First try conflict vars, then other vars.
        for var in conflict_vars + other_vars:
            successors = graph.var_post_out(var, reach_set)
            if not successors.is_empty():
                reach_set = reach_set.union(successors)
                all_done = False

                # This is a bit wasteful but at this point it
                # should be irrelevant for performance.
                if var in conflict_vars:
                    conflict_vars.remove(var)

                if var in other_vars:
                    other_vars.remove(var)

                saturated_vars.append(var)
                saturated_vars = sort_variable_list(saturated_vars)

                if sd.config["debug"]:
                    print(
                        f"[{node_id}] > Saturation({len(saturated_vars)}) Added saturation variable. {len(conflict_vars)} conflict and {len(other_vars)} other variables remaining."
                    )
                break

    if sd.config["debug"]:
        print(f"[{node_id}] > Reachability completed with {reach_set}.")

    return reach_set


def sort_variable_list(variables: list[VariableId]):
    return list(sorted(variables, reverse=True))
